@@ -25,8 +25,41 @@ def as_tuple(obj):
         return (obj,)
 
 
+def minmax_schema_rule(widget, schema):
+    """This implements the minimum/maximum rules
+
+    Only used for inputs bounded from one side, as ipywidgets
+    has dedicated widgets for inputs bound from both sides. Defined
+    in a separate function, because it used twice for number and
+    integer widgets.
+    """
+    if "minimum" in schema:
+
+        def min_handler(_):
+            if widget.value < schema["minimum"]:
+                widget.value = schema["minimum"]
+
+        widget.observe(min_handler, names="value")
+
+    if "maximum" in schema:
+
+        def max_handler(_):
+            if widget.value > schema["maximum"]:
+                widget.value = schema["maximum"]
+
+        widget.observe(max_handler, names="value")
+
+    return widget
+
+
 class Form:
-    def __init__(self, schema, vertically_place_labels=False, use_sliders=False):
+    def __init__(
+        self,
+        schema,
+        vertically_place_labels=False,
+        use_sliders=False,
+        preconstruct_array_items=0,
+    ):
         """Create a form with Jupyter widgets from a JSON schema
 
         :param schema:
@@ -42,6 +75,11 @@ class Form:
         :param use_sliders:
             Whether bounded input fields should use sliders or regular input widgets
         :type use_sliders: bool
+        :param preconstruct_array_items:
+            How many array item widgets should be pre-generated. For very complex
+            subschemas for each array item, this can be a performance optimization
+            trading construction time vs. usage delay.
+        :type preconstruct_array_items: int
         """
         # Make sure that the given schema is valid
         filename = os.path.join(
@@ -56,6 +94,7 @@ class Form:
         self.schema = schema
         self.vertically_place_labels = vertically_place_labels
         self.use_sliders = use_sliders
+        self.preconstruct_array_items = preconstruct_array_items
 
         # Store a list of registered observers to add them to runtime-generated widgets
         self._observers = []
@@ -215,6 +254,11 @@ class Form:
         # Apply a potential default
         if "default" in schema:
             widget.value = schema["default"]
+        else:
+            if "minimum" in schema:
+                widget.value = schema["minimum"]
+            if "maximum" in schema:
+                widget.value = schema["maximum"]
 
         # Apply potential constant values without generating a widget
         if "const" in schema:
@@ -285,7 +329,8 @@ class Form:
                 label=label,
             )
         else:
-            return self._construct_simple(schema, ipywidgets.FloatText(), label=label)
+            widget = minmax_schema_rule(ipywidgets.FloatText(), schema)
+            return self._construct_simple(schema, widget, label=label)
 
     def _construct_integer(self, schema, label=None, root=False):
         # Inputs bounded only from below or above are currently not supported
@@ -300,7 +345,8 @@ class Form:
                 label=label,
             )
         else:
-            return self._construct_simple(schema, ipywidgets.IntText(), label=label)
+            widget = minmax_schema_rule(ipywidgets.IntText(), schema)
+            return self._construct_simple(schema, widget, label=label)
 
     def _construct_boolean(self, schema, label=None, root=False):
         # Extract the labelling for the checkbox
@@ -336,11 +382,6 @@ class Form:
         # recursive implementation of this array element
         elements = []
 
-        # The subwidgets used for each element of the array. This differs
-        # from the widgets of above elements, because it includes added
-        # array controls like move up/down and delete
-        element_widgets = []
-
         # We separately to store the size of the elements array. The reason
         # for this is that we do want to minimize the amount of element creations
         # as these are very costly for complex schemas
@@ -481,8 +522,9 @@ class Form:
         button.on_click(add_entry)
 
         # Initialize the widget with the minimal number of subwidgets
-        for _ in range(schema.get("minItems", 0)):
+        for _ in range(max(schema.get("minItems", 0), self.preconstruct_array_items)):
             add_entry(_)
+        element_size = schema.get("minItems", 0)
 
         # If this is not the root document, we wrap this in an Accordion widget
         wrapped_vbox = [vbox]
